package build

import (
	"fmt"
	"strconv"

	"git.akyoto.dev/cli/q/src/build/asm"
	"git.akyoto.dev/cli/q/src/build/cpu"
	"git.akyoto.dev/cli/q/src/build/expression"
	"git.akyoto.dev/cli/q/src/build/fs"
	"git.akyoto.dev/cli/q/src/build/token"
	"git.akyoto.dev/cli/q/src/errors"
	"git.akyoto.dev/go/color/ansi"
)

// Function represents a function.
type Function struct {
	Name      string
	File      *fs.File
	Head      token.List
	Body      token.List
	Variables map[string]*Variable
	Assembler asm.Assembler
	CPU       cpu.CPU
	Error     error
}

// Compile turns a function into machine code.
func (f *Function) Compile() {
	f.Assembler.Label(f.Name)
	err := f.CompileTokens(f.Body)

	if err != nil {
		f.Error = err
		return
	}

	f.Assembler.Return()
}

// CompileTokens compiles a token list.
func (f *Function) CompileTokens(body token.List) error {
	start := 0
	groupLevel := 0
	blockLevel := 0

	for i, t := range body {
		if start == i && t.Kind == token.NewLine {
			start = i + 1
			continue
		}

		switch t.Kind {
		case token.NewLine:
			if groupLevel > 0 || blockLevel > 0 {
				continue
			}

			if start != -1 {
				instruction := body[start:i]
				err := f.CompileInstruction(instruction)

				if err != nil {
					return err
				}

				start = -1
			}

			start = i + 1

		case token.GroupStart:
			groupLevel++

		case token.GroupEnd:
			groupLevel--

		case token.BlockStart:
			blockLevel++

		case token.BlockEnd:
			blockLevel--
		}
	}

	return nil
}

// CompileInstruction compiles a single instruction.
func (f *Function) CompileInstruction(line token.List) error {
	if len(line) == 0 {
		return nil
	}

	if line[0].Kind == token.Keyword {
		return f.CompileKeyword(line)
	}

	expr := expression.Parse(line)

	if expr == nil {
		return nil
	}

	defer expr.Close()

	if isVariableDefinition(expr) {
		return f.CompileVariableDefinition(expr)
	}

	if isFunctionCall(expr) {
		return f.CompileFunctionCall(expr)
	}

	return errors.New(&errors.InvalidInstruction{Instruction: expr.Token.Text()}, f.File, expr.Token.Position)
}

// CompileKeyword compiles an instruction that starts with a keyword.
func (f *Function) CompileKeyword(line token.List) error {
	switch line[0].Text() {
	case "return":
		if len(line) > 1 {
			value := expression.Parse(line[1:])
			defer value.Close()
			// TODO: Set the return value
		}

		f.Assembler.Return()

	case "loop":
		blockStart := line.IndexKind(token.BlockStart) + 1
		blockEnd := line.LastIndexKind(token.BlockEnd)

		if blockStart == -1 {
			return errors.New(errors.MissingBlockStart, f.File, line[0].End())
		}

		if blockEnd == -1 {
			return errors.New(errors.MissingBlockEnd, f.File, line[len(line)-1].End())
		}

		return f.CompileTokens(line[blockStart:blockEnd])

	default:
		return errors.New(&errors.KeywordNotImplemented{Keyword: line[0].Text()}, f.File, line[0].Position)
	}

	return nil
}

// CompileVariableDefinition compiles a variable definition.
func (f *Function) CompileVariableDefinition(expr *expression.Expression) error {
	if len(expr.Children) < 2 {
		return errors.New(errors.MissingAssignValue, f.File, expr.LastChild().Token.End())
	}

	name := expr.Children[0].Token.Text()

	if f.identifierExists(name) {
		return errors.New(&errors.VariableAlreadyExists{Name: name}, f.File, expr.Children[0].Token.Position)
	}

	value := expr.Children[1]

	err := value.EachLeaf(func(leaf *expression.Expression) error {
		if leaf.Token.Kind == token.Identifier && !f.identifierExists(leaf.Token.Text()) {
			return errors.New(&errors.UnknownIdentifier{Name: leaf.Token.Text()}, f.File, leaf.Token.Position)
		}

		return nil
	})

	if err != nil {
		return err
	}

	reg, exists := f.CPU.FindFree()

	if !exists {
		panic("no free registers")
	}

	f.ExpressionToRegister(value, reg)
	f.CPU.Use(reg)

	f.Variables[name] = &Variable{
		Name:     name,
		Register: reg,
	}

	return nil
}

// CompileFunctionCall compiles a function call.
func (f *Function) CompileFunctionCall(expr *expression.Expression) error {
	funcName := expr.Children[0].Token.Text()
	parameters := expr.Children[1:]

	for i, parameter := range parameters {
		err := f.ExpressionToRegister(parameter, f.CPU.Syscall[i])

		if err != nil {
			return err
		}
	}

	if funcName == "syscall" {
		f.Assembler.Syscall()
	} else {
		f.Assembler.Call(funcName)
	}

	return nil
}

// ExpressionToRegister moves the result of an expression into the given register.
func (f *Function) ExpressionToRegister(root *expression.Expression, register cpu.Register) error {
	if root.IsLeaf() {
		return f.TokenToRegister(root.Token, register)
	}

	return errors.New(errors.NotImplemented, f.File, root.Token.Position)
}

// TokenToRegister moves a token into a register.
// It only works with identifiers, numbers and strings.
func (f *Function) TokenToRegister(t token.Token, register cpu.Register) error {
	switch t.Kind {
	case token.Identifier:
		name := t.Text()
		variable, exists := f.Variables[name]

		if !exists {
			return errors.New(&errors.UnknownIdentifier{Name: name}, f.File, t.Position)
		}

		f.Assembler.MoveRegisterRegister(register, variable.Register)
		return nil

	case token.Number:
		value := t.Text()
		n, err := strconv.Atoi(value)

		if err != nil {
			return err
		}

		f.Assembler.MoveRegisterNumber(register, uint64(n))
		return nil

	case token.String:
		return errors.New(errors.NotImplemented, f.File, t.Position)

	default:
		return errors.New(errors.InvalidExpression, f.File, t.Position)
	}
}

// PrintAsm shows the assembly instructions.
func (f *Function) PrintAsm() {
	ansi.Bold.Println(f.Name)
	ansi.Dim.Println("╭────────────────────────────────────────────────────────────")

	for _, x := range f.Assembler.Instructions {
		ansi.Dim.Print("│ ")
		fmt.Print(x.Mnemonic.String())

		if x.Data != nil {
			fmt.Print(" " + x.Data.String())
		}

		fmt.Print("\n")
	}

	ansi.Dim.Println("╰────────────────────────────────────────────────────────────")
}

// String returns the function name.
func (f *Function) String() string {
	return f.Name
}

// identifierExists returns true if the identifier has been defined.
func (f *Function) identifierExists(name string) bool {
	_, exists := f.Variables[name]
	return exists
}

// isVariableDefinition returns true if the expression is a variable definition.
func isVariableDefinition(expr *expression.Expression) bool {
	return expr.Token.Kind == token.Operator && expr.Token.Text() == ":="
}

// isFunctionCall returns true if the expression is a function call.
func isFunctionCall(expr *expression.Expression) bool {
	return expr.Token.Kind == token.Operator && expr.Token.Text() == "λ"
}
